(* Split Zip files

     Dr. J. Rathlev, D-24222 Schwentinental (kontakt(a)rathlev-home.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   J. Rathlev, Nov. 2012
   last changes: March 2018  (Delphi 10)

   Command line options:
   SplitZip <ZipName> [Directory] [Size] [/quiet]
     ZipName   : Name of Zip file to split
     Directory : Location for splitted Zip files (default: Temp)
     Size      : Maximum size of each splitted file (default: 700 MB)
     /quiet    : do not show progess status

   ExitCode;
     = 0 : ok
     = 1 : Zip file bnot found
     = 2 : Target directory could not be created
     = 3 : At least one file in the archive is larger than the maximum size
   *)

unit SplitZipMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.Buttons,
  Vcl.Samples.Spin, Vcl.ComCtrls, ExtFileTools, NumberEd, Vcl.ExtCtrls;

const
  ProgName = 'Split Zip Files';
  Vers = '2.6';
  CopRgt = ' 2012-2020 Dr. J. Rathlev, 24222 Schwentinental';
  EmailAdr = 'kontakt(a)rathlev-home.de';

  ZipExt = 'zip';
  SpName = 'SplitFile';
  BlockSize = 256*1024;
  MB = 1024*1024;
  QuietOption = '/quiet';
    
type
  TZipFileInfo = record
    Filename        : string;
    Flag,Method     : word;
    TimeStamp,Crc   : cardinal;
    CSize,USize     : int64;
    PkExtraCrypt    : TPKExtraEncField;
    end;

  THauptForm = class(TForm)
    edZipFile: TEdit;
    Label1: TLabel;
    FileBtn: TSpeedButton;
    DestSelectBtn: TSpeedButton;
    Label2: TLabel;
    edDestDir: TEdit;
    btnConvert: TSpeedButton;
    InfoBtn: TSpeedButton;
    EndeBtn: TBitBtn;
    OpenDialog: TOpenDialog;
    pbSplits: TProgressBar;
    sbStatus: TStatusBar;
    GroupBox1: TGroupBox;
    rbCD: TRadioButton;
    rbDVD: TRadioButton;
    rbOther: TRadioButton;
    laMaxSize: TLabel;
    edMaxSize: TEdit;
    pnSettings: TPanel;
    procedure InfoBtnClick(Sender: TObject);
    procedure EndeBtnClick(Sender: TObject);
    procedure FileBtnClick(Sender: TObject);
    procedure DestSelectBtnClick(Sender: TObject);
    procedure btnConvertClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure rbCDClick(Sender: TObject);
    procedure rbDVDClick(Sender: TObject);
    procedure rbOtherClick(Sender: TObject);
    procedure edMaxSizeChange(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormActivate(Sender: TObject);
  private
    { Private-Deklarationen }
    ProgVersName,
    ProgVersDate,
    UserPath,AppPath : string;
    FileList    : TStringList;
    MaxSize     : int64;
    Quiet,
    Remote      : boolean;
    function ReadSize : int64;
    function ErrorText (ec : Integer) : string;
    function SplitZip (FName,DDir : string; MSize : int64) : integer;
  public
    { Public-Deklarationen }
    function RunQuiet : boolean;
  end;

var
  HauptForm: THauptForm;

implementation

{$R *.dfm}

uses ShellDirDlg, WinUtils, GnuGetText, LangUtils, FileUtils, FileConsts,
  NumberUtils, WinApiUtils, WinShell, Winapi.ShlObj, InitProg;

{ ------------------------------------------------------------------- }
procedure THauptForm.FormCreate(Sender: TObject);
var
  n  : integer;
begin
  TranslateComponent(self);
  InitPaths(AppPath,UserPath);
  InitVersion(_('Split a Zip file'),Vers,CopRgt,3,3,ProgVersName,ProgVersDate);
  Caption:=_('Split a Zip file')+' ('+VersInfo.Comments+')';
  FileList:=TStringList.Create;
  FileList.Sorted:=false;
  MaxSize:=700*MB;
  // Befehlszeile einlesen
  Remote:=ParamCount>0; Quiet:=false;
  if Remote then begin
    edDestDir.Text:=TempDirectory;
    edZipFile.Text:=ParamStr(1);
    if ParamCount>1 then begin
      if AnsiSameText(ParamStr(2),QuietOption) then Quiet:=true
      else begin
        edDestDir.Text:=ParamStr(2);
        if ParamCount>2 then begin
          if AnsisameText(ParamStr(3),QuietOption) then Quiet:=true
          else if TryStrToInt(ParamStr(3),n) then MaxSize:=n*int64(MB);
          if (ParamCount>3) and AnsisameText(ParamStr(4),QuietOption) then Quiet:=true;
          end;
        end;
      end;
    end;
  end;

procedure THauptForm.FormDestroy(Sender: TObject);
var
  i : integer;
begin
  with FileList do begin
    for i:=0 to Count-1 do Objects[i].Free;
    Free;
    end;
  end;

procedure THauptForm.FormShow(Sender: TObject);
begin
  if Remote then begin
    pnSettings.Visible:=false;
    FileBtn.Visible:=false;
    DestSelectBtn.Visible:=false;
    ClientHeight:=ClientHeight-101;
    ClientWidth:=ClientWidth-31;
    end;
  end;

procedure THauptForm.FormActivate(Sender: TObject);
begin
  if Remote then begin
    ExitCode:=SplitZip(edZipFile.Text,edDestDir.Text,MaxSize);
    Application.Terminate;
    end;
  end;

function THauptForm.RunQuiet : boolean;
begin
  Result:=Quiet;
  if Quiet then begin
    Visible:=false;
    ExitCode:=SplitZip(edZipFile.Text,edDestDir.Text,MaxSize);
    end;
  end;

procedure THauptForm.InfoBtnClick(Sender: TObject);
begin
  InfoDialog (ProgName,Caption+' - '+ProgVersDate+sLineBreak+
           VersInfo.CopyRight);
  end;

procedure THauptForm.rbCDClick(Sender: TObject);
begin
  MaxSize:=700*int64(MB);
  edMaxSize.Enabled:=false;
  laMaxSize.Enabled:=false;
  end;

procedure THauptForm.rbDVDClick(Sender: TObject);
begin
  MaxSize:=4700*int64(MB);
  edMaxSize.Enabled:=false;
  laMaxSize.Enabled:=false;
  end;

function THauptForm.ReadSize : int64;
var
  v : double;
begin
  if TryStrToFloat(edMaxSize.Text,v) then Result:=round(v*MB)
  else begin
    ErrorDialog(_('Invalid size specification')); //'Ungltige Grenangabe!'
    Result:=MaxSize;
    end;
  end;

procedure THauptForm.rbOtherClick(Sender: TObject);
begin
  MaxSize:=ReadSize;
  edMaxSize.Enabled:=true;
  laMaxSize.Enabled:=true;
  edMaxSize.SetFocus;
  end;

procedure THauptForm.edMaxSizeChange(Sender: TObject);
begin
  MaxSize:=ReadSize;
  end;

procedure THauptForm.EndeBtnClick(Sender: TObject);
begin
  Close;
  end;

procedure THauptForm.FileBtnClick(Sender: TObject);
begin
  with OpenDialog do begin
    if length(edZipFile.Text)>0 then InitialDir:=ExtractFilePath(edZipFile.Text)
    else InitialDir:=UserPath;
    if Execute then edZipFile.Text:=Filename;
    end;
  end;

procedure THauptForm.DestSelectBtnClick(Sender: TObject);
var
  s : string;
begin
  s:=ExtractFilePath(edZipFile.Text);
  if length(s)=0 then s:=UserPath;
  if ShellDirDialog.Execute (_('Select destination directory'),false,true,false,'',s) then
    edDestDir.Text:=s;  //'Zielverzeichnis auswhlen'
  end;

// Result = 0 : ok
//          1 : Zip-Datei nicht gefunden
//          2 : Zielverzeichnis konnte nicht erstellt werden
//          3 : zu groe Einzeldateien
function THauptForm.SplitZip (FName,DDir : string; MSize : int64) : integer;
var
  RZip        : TZipRead;
  WZip        : TZipWrite;
  bt,bp,
  fp,bc       : int64;
  mfc,i,ns,
  fc,j        : integer;
  FileSizes   : array of int64;
  ok          : boolean;
  SplitName,s : string;
begin
  if not FileExists(FName) then begin
    Result:=1; Exit;
    end;
  RZip:=TZipRead.Create(FName,'',0,defBufferSize);
  RZip.ReadDirectory;
  SplitName:=ExtractFilename(FName);
  if not Quiet then begin
    pbSplits.Position:=0;
    Application.ProcessMessages;
    end;
  mfc:=0; ok:=true;
  MSize:=MSize-sizeof(TPKEndHeader)-sizeof(TPKZip64CdRecord)-sizeof(TPKZip64CdLocator);
  SetLength(FileSizes,1); FileSizes[0]:=0;
  with RZip,Directory do for i:=0 to Count-1 do with (Objects[i] as TPkHeader) do if not IsDir then begin
    if i=Count-1 then bc:=DirectoryOffset-Offset
    else bc:=(Objects[i+1] as TPkHeader).Offset-Offset;
    inc(bc,DirLength);
    if bc>MSize then begin
      ok:=false; Break;
      end;
    fc:=-1;
    for j:=0 to mfc do if FileSizes[j]+bc<MSize then begin
      fc:=j; Break;
      end;
    if fc<0 then begin
      inc(mfc); SetLength(FileSizes,mfc+1);
      FileSizes[mfc]:=0;
      fc:=mfc;
      end;
    inc(FileSizes[fc],bc);
    Tag:=fc;
    end;
  bt:=0; bp:=0;
  for j:=0 to mfc do bt:=bt+FileSizes[j];
  FileSizes:=nil;
  if ok then begin // Starte Splitten
    if ForceDirectories(DDir) then begin
      ns:=NumberOfDigits(mfc);
      for j:=0 to mfc do begin
        s:=InsertNameSuffix(SplitName,'-'+ZStrInt(j,ns));
        WZip:=TZipWrite.Create(SetDirName(DDir)+s,'','',0,defBufferSize);
        sbStatus.SimpleText:=_(' Writing: ')+s;
        with RZip,Directory do for i:=0 to Count-1 do with (Objects[i] as TPkHeader) do
            if not IsDir and (Tag=j) then begin
          if i=Count-1 then bc:=DirectoryOffset-Offset
          else bc:=(Objects[i+1] as TPkHeader).Offset-Offset;
          FilePosition:=Offset;
          fp:=WZip.FilePosition;
          bp:=bp+bc;
          CopyToZip(WZip,bc);
          WZip.Directory.AddObject(Strings[i],TPkHeader.Assign(Objects[i] as TPkHeader,fp));
          if not Quiet then begin
            pbSplits.Position:=round(100/bt*bp);
            Application.ProcessMessages;
            end;
          end;
        WZip.WriteDirectory;
        WZip.Free;
        end;
      Result:=0;
      end
    else Result:=2;
    end
  else Result:=3;
  if Result=0 then sbStatus.SimpleText:=Format(_('Done (%s)'),[ErrorText(Result)])
  else sbStatus.SimpleText:=Format(_('Error: %s!'),[ErrorText(Result)]);
  RZip.Free;
  end;

function THauptForm.ErrorText (ec : Integer) : string;
begin
  case ec of
  1: Result:=_('Zip file not found');
  2: Result:=_('Could not create destination directory');
  3: Result:=_('Size of single file exceeds given maximum');
  else Result:=_('No errors');
    end;
  end;

procedure THauptForm.btnConvertClick(Sender: TObject);
var
  n : integer;
begin
  n:=SplitZip(edZipFile.Text,edDestDir.Text,MaxSize);
  case n of
  1 : ErrorDialog(ErrorText(n)+': '+edZipFile.Text); //'Zip-Datei nicht gefunden: '
  2 : ErrorDialog(ErrorText(n)+': '+edDestDir.Text); //'Verzeichnis konnte nicht erstellt werden: '
  3 : ErrorDialog(ErrorText(n)+_(' - no splitting possible!')); //'Einzeldatei grer als Maximum - Kein Aufspalten mglich!'
    end;
  end;

end.
